/* $Id: vga_modex.inc,v 1.5 1998/11/23 20:37:12 marcus Exp $
******************************************************************************

   IBM VGA ModeX accelerations driver

   Copyright (C) 1996-1997 Jason McMullan	[jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************

   This code was derived from the following sources of information:
   [1] Xfree86 VGA driver

******************************************************************************
*/

#define XVIRTUAL current_mode.virt.x
#define YVIRTUAL current_mode.virt.y

#define CHECKXY {if (x<0 || y<0) return -ENBOGUS; \
                 if (x >= XVIRTUAL || y >= YVIRTUAL) return -EOK;}
#define CHECKXYH {if (h<0) return -ENBOGUS; if (h==0) return -EOK; \
                  CHECKXY; if ((y+h) > YVIRTUAL) h=YVIRTUAL-y;}
#define CHECKXYW {if (w<0) return -ENBOGUS; if (w==0) return -EOK; \
                  CHECKXY; if ((x+w) > XVIRTUAL) w=XVIRTUAL-x;}
#define CHECKXYWH {CHECKXY; if (w<0 || h<0) return -ENBOGUS; \
                   if (w==0 || h==0) return -EOK; \
                   if ((x+w) > XVIRTUAL) w=XVIRTUAL-x; \
                   if ((y+h) > YVIRTUAL) h=YVIRTUAL-y;}
                   

/* NOTE: this doubles as putpixel */
int drawpixel(int x,int y,int color)
{
	unsigned int ndx;

	CHECKXY;

	ndx = (x + y * XVIRTUAL) >> 2;
	outbSEQ(1 << (x & 3), 0x02);
	*((char *)fb_graphics + ndx) = color;
	return -EOK;
}

#define putpixel(x,y,col) drawpixel(x,y,col)

int drawvline(int x, int y, int h, int color)
{
	unsigned int ndx;
	int i;
	
	CHECKXYH;

	ndx = (x + y*XVIRTUAL) >> 2;
	outbSEQ(1 << (x & 3), 0x02);
	for (i = 0; i < h; i++) {
		*((char *)fb_graphics + ndx) = color;
		ndx += XVIRTUAL >> 2;
	}

	return -EOK;
}

int putvline(int x, int y, int h, void *content)
{
	unsigned int ndx;
	int i;

	/* Catch stupidity */
	CHECKXYH;

	if (verify_area(VERIFY_READ, content, h) != 0)
		return -ENBOGUS;

	ndx = (x + y * XVIRTUAL) >> 2;
	outbSEQ(1 << (x & 3), 0x02);
	for (i = 0; i < h; i++) {
		*((char *)fb_graphics+ndx) = *(((char *)content) + i);
		ndx += XVIRTUAL >> 2;
	}
	
	return -EOK;
}

/* Hlines */

int drawhline(int x, int y, int w, int color)
{
	unsigned int ndx;
	int i, mask, len, start;

	CHECKXYW;

	/* Draw the 'front' pixels */
	if ((x&0x3) != 0) {
		mask = 0;
		for (i = (x & 3); i < 4 && w != 0; i++, w--, x++) 
			mask |= (1 << (x & 3));

		outbSEQ(mask, 0x02);
		ndx = ((x - 1) + y * XVIRTUAL) >> 2;
		((char *)fb_graphics)[ndx] = (char)(color);
	}

	if (w == 0)
		return -EOK;

	if ((x & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}

	/* Draw the 'back' pixels */
	if (((x + w) & 0x3) != 0) {
		mask = 0x0f;
		for (i = ((x + w) & 3); i < 4; i++) 
			mask ^= (1 << i);

		outbSEQ(mask, 0x02);
		ndx = ((x + w) + y * XVIRTUAL) >> 2;
		((char *)fb_graphics)[ndx] = (char)(color);

		w = ((x + w) & (~0x03)) - x;
	}
   
	if (w <= 0)
		return EOK;

	if (((x + w) & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}
  
	/* Fill in the rest of the line */
	outbSEQ(0x0f, 0x02);
	start=(x + y * XVIRTUAL) >> 2;
	len = w >> 2;
	memset(((char *)fb_graphics) + start, color, len);

	return -EOK;
}

int puthline(int x, int y, int w, char *colors)
{
	unsigned int ndx;
	int i, mask, pixel;
	int ox, oy, ow;

	CHECKXYW;
	
	ox = x; oy = y; ow = w;

	if (verify_area(VERIFY_READ, colors, w) != 0)
		return -ENBOGUS;

	pixel = 0;
	/* Draw the 'front' pixels */
	if ((x & 0x3) != 0) {
		mask = 0;
		ndx = (x + y * XVIRTUAL) >> 2;
		for (i = (x & 3); i < 4 && w != 0; i++, w--, x++) {
			mask = (1 << (x & 3));
			outbSEQ(mask, 0x02);
			((char *)fb_graphics)[ndx] = 
				get_user(colors + (pixel++));
		}
	}

	if (w <= 0)
		return -EOK;

	if ((x & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}

	/* Draw the 'back' pixels */
	w--; /* Base 0 now */

	if (((x + w) & 0x3) != 0x3) {
		ndx = ((x + w) + y * XVIRTUAL) >> 2;
		for (i = ((x + w) & 3); i >= 0; i--)  {
			mask = (1 << i);
			outbSEQ(mask, 0x02);
			((char *)fb_graphics)[ndx] = get_user(colors + (w--));
		}
	}
	
	w++; /* Back to base 1 */
   
	if (w <= 0)
		return EOK;

	if (((x + w) & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}
  
	/* Fill in the rest of the line */
	colors += pixel;

	for (i = 0; i < 4; i++) {
		ndx = (x + y * XVIRTUAL) >> 2;
		mask = 1 << i;
		outbSEQ(mask, 0x02);
		for (pixel = i; pixel < w; pixel += 4, ndx++) {
			((char *)fb_graphics)[ndx] = get_user(colors + pixel);
		}
	} 

	return -EOK;
}

/* Boxes */
int drawbox(int x, int y, int w, int h, int color)
{
	unsigned int ndx;
	int i, mask, len, start;

	CHECKXYWH;

	/* Draw the 'front' lines */
	if ((x & 0x3) != 0) {
		mask = 0;
		for (i = (x & 3); i < 4 && w != 0; i++, w--, x++) 
			mask |= (1 << (x & 3));

		outbSEQ(mask, 0x02);
		ndx = ((x - 1) + y * XVIRTUAL) >> 2;
		for (i = 0; i < h; i++) {
			((char *)fb_graphics)[ndx] = (char)(color);
			ndx += (XVIRTUAL >> 2);
		}
	}

	if (w == 0)
		return -EOK;

	if ((x & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}

	/* Draw the 'back' lines */
	if (((x + w) & 0x3) != 0) {
		mask = 0x0f;
		for (i = ((x + w) & 3); i < 4; i++) 
			mask ^= (1 << i);

		outbSEQ(mask, 0x02);
		ndx = ((x + w) + y * XVIRTUAL) >> 2;
		for (i = 0; i < h; i++) {
			((char *)fb_graphics)[ndx] = (char)(color);
			ndx += (XVIRTUAL >> 2);
		}

		w = ((x + w) & (~0x3)) - x;
	}
   
	if (w <= 0)
		return EOK;

	if (((x + w) & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_ALWAYS_LOWER;
	}
  
	/* Fill in the rest of the box */
	outbSEQ(0x0f, 0x02);
	start = (x + y * XVIRTUAL) >> 2;
	len = w >> 2;
	ndx = XVIRTUAL >> 2;
	for (i = 0; i < h; i++) {
		memset(((char *)fb_graphics) + start, color, len);
		start += ndx;
    }

	return -EOK;
}

int putbox(int x, int y, int w, int h, char *colors)
{
	unsigned int ndx;
	int i, j, mask, pixel;
	int ow;

	CHECKXYWH;

	ow = w;

	if (verify_area(VERIFY_READ, colors, w * h) != 0)
		return -ENBOGUS;

	pixel = 0;
	/* Draw the 'front' pixels */
	if ((x & 0x3) != 0) {
		mask = 0;
		ndx = (x + y * XVIRTUAL) >> 2;
		for (i = (x & 3); i < 4 && w != 0; i++, w--, x++) {
			unsigned int ptr, nptr;

			mask = (1 << (x & 3));
			outbSEQ(mask, 0x02);
			ptr = pixel;
			nptr = ndx;
			for (j = 0; j < h; j++) {
				((char *)fb_graphics)[nptr] = 
					get_user(colors + ptr);
				ptr += ow;
				nptr += XVIRTUAL >> 2;
			}
			pixel++;
		}
	}

	if (w <= 0)
		return -EOK;

	if ((x & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}

	/* Draw the 'back' pixels */
	w--; /* Base 0 now */

	if (((x + w) & 0x3) != 0x3) {
		ndx = ((x + w) + y * XVIRTUAL) >> 2;
		for (i = ((x + w) & 3); i >= 0; i--)
		{
			unsigned int ptr, nptr;

			mask = (1 << i);
			outbSEQ(mask, 0x02);
			ptr = w;
			nptr = ndx;
			for (j = 0; j < h; j++)
			{
				((char *)fb_graphics)[nptr] = 
					get_user(colors + ptr);
				ptr += ow;
				nptr += XVIRTUAL >> 2;
			}
			w--;
		}
	}
	w++; /* Back to base 1 */
   
	if (w <= 0)
		return EOK;

	if (((x + w) & 3) != 0) {
		INTERNAL_ERROR;
		return -ENODRVSUP_MODE_LOWER;
	}
  
	/* Fill in the rest of the line */
	colors += pixel;
	
	for (i = 0; i < 4; i++) {
		unsigned int ptr, nptr;

		ndx = (x + y * XVIRTUAL) >> 2;
		mask = 1 << i;
		outbSEQ(mask, 0x02);
		for (pixel = i; pixel < w; pixel += 4, ndx++) {
			ptr = pixel;
			nptr = ndx;
			for (j = 0; j < h; j++) {
				((char *)fb_graphics)[nptr] =
					get_user(colors + ptr);
				ptr += ow;
				nptr += XVIRTUAL >> 2;
			}
		}
	} 
	
	return -EOK;
}

/* Fill screen */

int fillscreen(int color)
{
	outbSEQ(0x0f, 0x02);
	memset(fb_graphics, color, fb_graph_len);
	return -EOK;
}
